/**
 *@file ftds_decode.h
 *
 *@brief This file contains function prototypes and data definitions that
 *helps to decode FTDS frame and storing data
 */

#ifndef FTDP_DECODE_H_
#define FTDP_DECODE_H_

#include "data_type.h"
#include "notify_apps.h"


/*!
 * @brief This API receives the FTDS frame data from TC-TM then decode FTDS frame
 *
 */
void ft_receive_ftds(uint8_t *ftds_data, uint16_t ftds_data_length);






/* internal FT functions */
void decode_ftds(void);
void ft_seq_loss_handle_init(void);
uint8_t ftds_decode(void);
uint8_t prepare_sts_and_notify_receiver_app(ft_status_type status);
void save_context_info(ft_notification_info* sts);
uint32_t ft_get_receiver_head_seq_id(void);
uint8_t ft_get_seq_loss_sts(void);
void ft_set_app_id_rx(uint16_t set);
uint16_t ft_get_app_id_rx(void);
void ft_ftds_busy_clear(void);
void ft_clear_rec_retranmission_status(void);
uint32_t ft_get_rec_retranmission_status(void);



#endif /*FTDP_DECODE_H_ */
